package cs108;

import java.awt.image.BufferedImage;

/**
 * Une image continue et infinie, représentée par une fonction associant une
 * couleur à chaque point du plan.
 */
@FunctionalInterface
public interface ImageRGB {
    public ColorRGB apply(double x, double y);

    public static final ImageRGB RED_CIRCLE = (x, y) -> {
        double r = Math.sqrt(x * x + y * y);
        return r <= 1.0 ? ColorRGB.RED : ColorRGB.WHITE;
    };

    public default ImageRGB flattened() {
        return (x, y) -> apply(x, y * 2d);
    }

    public static ImageRGB chessboard(ColorRGB c1, ColorRGB c2, double w) {
        if (! (w > 0))
            throw new IllegalArgumentException("non-positive width: " + w);

        return (x, y) -> {
            double sqX = Math.floor(x / w), sqY = Math.floor(y / w);
            return ((int)(sqX + sqY)) % 2 == 0 ? c1 : c2;
        };
    }

    public static ImageRGB target(ColorRGB c1, ColorRGB c2, double w) {
        if (! (w > 0))
            throw new IllegalArgumentException("non-positive width: " + w);

        return (x, y) -> {
            double r = Math.sqrt(x * x + y * y);
            return (int)Math.floor(r / w) % 2 == 0 ? c1 : c2;
        };
    }

    public static ImageRGB linearHorizontalGradient(ColorRGB cL, double xL, ColorRGB cR, double xR) {
        if (! (xR > xL))
            throw new IllegalArgumentException("invalid interval "+ xL + " - " + xR);

        return (x, y) -> {
            if (x <= xL)
                return cL;
            else if (x >= xR)
                return cR;
            else {
                double p = (x - xL) / (xR - xL);
                return cL.mixWith(cR, p);
            }
        };
    }

    public default ImageRGB rotated(double a) {
        double cosInvA = Math.cos(-a);
        double sinInvA = Math.sin(-a);
        return (x, y) -> {
            double rX = x * cosInvA - y * sinInvA;
            double rY = x * sinInvA + y * cosInvA;
            return apply(rX, rY);
        };
    }

    public static ImageRGB ofDiscreteImage(ColorRGB bg, BufferedImage i) {
        double w = i.getWidth();
        double h = i.getHeight();
        double s = w / 2d;

        return (x, y) -> {
            int iX = (int)Math.round( x * s + w / 2d);
            int iY = (int)Math.round(-y * s + h / 2d);

            if (0 <= iX && iX < w && 0 <= iY && iY < h)
                return new ColorRGB(i.getRGB(iX, iY));
            else
                return bg;
        };
    }
}
